@@ -32,6 +32,7 @@ class PlotGenerator():
 		constraint_eval_fns=[],
 		perf_eval_kwargs={},
 		constraint_eval_kwargs={},
+		batch_epoch_dict={}
 		):
 		""" Class for running Seldonian experiments 
 		and generating the three plots:
@@ -84,6 +85,10 @@ class PlotGenerator():
 		:param constraint_eval_kwargs: Extra keyword arguments to pass to
 			the constraint_eval_fns
 		:type constraint_eval_kwargs: dict
+
+		:param batch_epoch_dict: Instruct batch sizes and n_epochs
+			for each data frac
+		:type batch_epoch_dict: dict
 		"""
 		self.spec = spec
 		self.n_trials = n_trials
@@ -95,9 +100,11 @@ class PlotGenerator():
 		self.constraint_eval_fns = constraint_eval_fns
 		self.perf_eval_kwargs = perf_eval_kwargs
 		self.constraint_eval_kwargs = constraint_eval_kwargs
+		self.batch_epoch_dict = batch_epoch_dict
 
 	def make_plots(self,fontsize=12,legend_fontsize=8,
 		performance_label='accuracy',
+		performance_yscale='linear',
 		marker_size=20,
 		include_legend=True,
 		savename=None):
@@ -219,9 +226,7 @@ class PlotGenerator():
 
 			# Plot labels
 			ax_performance.set_ylabel(performance_label,fontsize=fontsize)
-			# ax_sr.set_ylabel('Solution rate',fontsize=fontsize)
 			ax_sr.set_ylabel('Probability of solution',fontsize=fontsize)
-			# ax_fr.set_ylabel('Failure Rate',fontsize=fontsize)
 			ax_fr.set_ylabel('Probability constraint was violated',fontsize=fontsize)
 
 			# Only put horizontal axis labels on last row of plots 
@@ -229,12 +234,14 @@ class PlotGenerator():
 				# ax_performance.set_xlabel('Training samples',fontsize=fontsize)
 				# ax_sr.set_xlabel('Training samples',fontsize=fontsize)
 				# ax_fr.set_xlabel('Training samples',fontsize=fontsize)
-				ax_performance.set_xlabel('Size of training set',fontsize=fontsize)
-				ax_sr.set_xlabel('Size of training set',fontsize=fontsize)
-				ax_fr.set_xlabel('Size of training set',fontsize=fontsize)
+				ax_performance.set_xlabel('Amount of data',fontsize=fontsize)
+				ax_sr.set_xlabel('Amount of data',fontsize=fontsize)
+				ax_fr.set_xlabel('Amount of data',fontsize=fontsize)
 
 			# axis scaling
 			ax_performance.set_xscale('log')
+			if performance_yscale.lower() == 'log':
+				ax_performance.set_yscale('log')
 			ax_sr.set_xscale('log')
 			ax_fr.set_xscale('log')
 
@@ -285,10 +292,9 @@ class PlotGenerator():
 				ste_performance = std_performance/np.sqrt(n_passed)
 				X_passed_seldonian = this_seldonian_dict['X_passed']
 				pl, = ax_performance.plot(X_passed_seldonian,mean_performance,color=seldonian_color,
-					linestyle='--',label='QSA')
+					linestyle='--')
 				legend_handles.append(pl)
 				legend_labels.append(seldonian_model)
-				# legend_labels.append('Linear regression with constraint')
 				ax_performance.scatter(X_passed_seldonian,mean_performance,color=seldonian_color,
 					s=marker_size,marker='o')
 				ax_performance.fill_between(X_passed_seldonian,
@@ -401,7 +407,6 @@ class PlotGenerator():
 		else:
 			plt.show()
 
-
 class SupervisedPlotGenerator(PlotGenerator):
 	def __init__(self,
 		spec,
@@ -414,6 +419,7 @@ class SupervisedPlotGenerator(PlotGenerator):
 		constraint_eval_fns=[],
 		perf_eval_kwargs={},
 		constraint_eval_kwargs={},
+		batch_epoch_dict={},
 		):
 		"""Class for running supervised Seldonian experiments 
 			and generating the three plots
@@ -462,6 +468,10 @@ class SupervisedPlotGenerator(PlotGenerator):
 		:param constraint_eval_kwargs: Extra keyword arguments to pass to
 			the constraint_eval_fns
 		:type constraint_eval_kwargs: dict
+
+		:param batch_epoch_dict: Instruct batch sizes and n_epochs
+			for each data frac
+		:type batch_epoch_dict: dict
 		"""
 
 		super().__init__(spec=spec,
@@ -474,6 +484,7 @@ class SupervisedPlotGenerator(PlotGenerator):
 			constraint_eval_fns=constraint_eval_fns,
 			perf_eval_kwargs=perf_eval_kwargs,
 			constraint_eval_kwargs=constraint_eval_kwargs,
+			batch_epoch_dict=batch_epoch_dict,
 			)
 		self.regime = 'supervised_learning'
 
@@ -508,6 +519,7 @@ class SupervisedPlotGenerator(PlotGenerator):
 			perf_eval_kwargs=self.perf_eval_kwargs,
 			constraint_eval_fns=self.constraint_eval_fns,
 			constraint_eval_kwargs=self.constraint_eval_kwargs,
+			batch_epoch_dict=self.batch_epoch_dict,
 			verbose=verbose,
 			)
 
@@ -553,6 +565,7 @@ class SupervisedPlotGenerator(PlotGenerator):
 			perf_eval_kwargs=self.perf_eval_kwargs,
 			constraint_eval_fns=self.constraint_eval_fns,
 			constraint_eval_kwargs=self.constraint_eval_kwargs,
+			batch_epoch_dict=self.batch_epoch_dict,
 			verbose=verbose,
 			)
 
@@ -631,6 +644,7 @@ class RLPlotGenerator(PlotGenerator):
 		constraint_eval_fns=[],
 		perf_eval_kwargs={},
 		constraint_eval_kwargs={},
+		batch_epoch_dict={},
 		):
 		"""Class for running RL Seldonian experiments 
 			and generating the three plots
@@ -679,6 +693,10 @@ class RLPlotGenerator(PlotGenerator):
 		:param constraint_eval_kwargs: Extra keyword arguments to pass to
 			the constraint_eval_fns
 		:type constraint_eval_kwargs: dict
+
+		:param batch_epoch_dict: Instruct batch sizes and n_epochs
+			for each data frac
+		:type batch_epoch_dict: dict
 		"""
 
 		super().__init__(spec=spec,
@@ -691,6 +709,7 @@ class RLPlotGenerator(PlotGenerator):
 			constraint_eval_fns=constraint_eval_fns,
 			perf_eval_kwargs=perf_eval_kwargs,
 			constraint_eval_kwargs=constraint_eval_kwargs,
+			batch_epoch_dict=batch_epoch_dict,
 			)
 		
 		self.regime = 'reinforcement_learning'
@@ -737,6 +756,7 @@ class RLPlotGenerator(PlotGenerator):
 			constraint_eval_kwargs=self.constraint_eval_kwargs,
 			perf_eval_fn=self.perf_eval_fn,
 			perf_eval_kwargs=self.perf_eval_kwargs,
+			batch_epoch_dict=self.batch_epoch_dict,
 			verbose=verbose,
 			)
 
